#!/usr/bin/env perl
#pk

##########################################################################################
#	  This file is part of proteinortho.
#	  (C) 2009 Marcus Lechner
# 
#	  proteinortho is free software; you can redistribute it and/or modify
#	  it under the terms of the GNU General Public License as published
#	  by the Free Software Foundation; either version 2, or (at your
#	  option) any later version.
#
#	  proteinortho is distributed in the hope that it will be useful, but
#	  WITHOUT ANY WARRANTY; without even the implied warranty of
#	  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#	  General Public License for more details.
#
#	  You should have received a copy of the GNU General Public License
#	  along with proteinortho; see the file COPYING.  If not, write to the
#	  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
#	  Boston, MA 02111-1307, USA.	
##########################################################################################

##########################################################################################
# About
##########################################################################################
# 
# @author Paul Klemm
# @email klemmp@staff.uni-marburg.de
# @company Bioinformatics, University of Leipzig
# @version 1
# @date 11-11-2019
#
##########################################################################################

use POSIX;

my $usage = <<'ENDUSAGE';
proteinortho_history.pl        reports the history of a (or a pair of) gene/protein(s).
 
SYNOPSIS
 
proteinortho_history.pl (-project=myproject) QUERY (FASTA1 FASTA2 ...)

	QUERY	A string of a single gene/protein or 2 separated by a comma or a whitespace (the input is escaped using quotemeta, use -noquotemeta to avoid this)

	-project=MYPROJECT	The project name (as specified in proteinortho with -project) (default:myproject)
	-step=[123] 		(optional) If specified more optput is printed (to STDOUT) for the given step:
		-step=1 : search for the given fasta sequence in the input fasta files
		-step=2 : search in the *.blast-graph
		-step=3 : search in the *.proteinortho file (default:nothing)
		-step=all : prints everything of above to STDOUT
	FASTA*				(optional) input fasta files 
	-noquotemeta 		(optional) If set, then the query will not be escaped.
	-delim= 		(optional) Defines the delimiter character for spliting the query (if you want to search for 2 genes/proteins)

	NOTE: if you use the -keep option and you have the project_cache_proteinortho/ directory, this program additionally searches for all blast hits.

ENDUSAGE

my $query;
my $help;
my $project="myproject";
my $step="none";
my $delim="[, ]";
my $do_quotemeta=1;

my @ARGViddone=[];
my $ARGViddone_counter=scalar(@ARGV);
for(my $v = 0 ; $v < scalar @ARGV ; $v++){
	
	if($ARGV[$v] =~ m/--?(help|h)$/){$help=1;$ARGViddone[$v]=1;}
	elsif($ARGV[$v] =~ m/^--?project=?(.*)$/){$project=$1;chomp($project);$ARGViddone[$v]=1;}
	elsif($ARGV[$v] =~ m/^--?step=?([123]|(all))$/){$step=$1;chomp($step);$ARGViddone[$v]=1;}
	elsif($ARGV[$v] =~ m/^--?delim=?(.*)$/){$delim=$1;chomp($delim);$ARGViddone[$v]=1;}
	elsif($ARGV[$v] =~ m/^--?noquotemeta$/){$do_quotemeta=0;$ARGViddone[$v]=1;}
	elsif($ARGV[$v] =~ m/^-.+/){print $usage; print STDERR "ERROR: invalid option ".$ARGV[$v]."!\n\n";$ARGViddone[$v]=1;exit(1);}
	elsif(!defined($query)){$query = $ARGV[$v];}
	else{$ARGViddone[$v]=0;$ARGViddone_counter--;}

}

if ($help){
    print $usage;
    exit(0);
}
my $fail="";
if (!defined($query)){
    $fail.="ERROR: no QUERY provided!\n";
}
if (! -e $project.".blast-graph" && ! -e $project.".proteinortho.tsv" && ! -e $project.".proteinortho-graph" && ! -d "proteinortho_cache_".$project){
    $fail.="ERROR: '".$project."' files not found, make sure you are in the directory with the proteinortho output! Specify the project name (prefix) with -project=... \n";
}

if($do_quotemeta){
	$query=quotemeta($query);
	$query=~s/\\([, ])/$1/;
	$query=~s/\\($delim)/$1/;
}

my @query_split=split($delim,$query);

if (scalar(@query_split) > 2){
    $fail.="ERROR: the QUERY contains too many delimiters ([, ]), specify the delimiter character with -delim=... !\n";
}

if($fail){
	print $usage.$fail;
	exit(1);
}

print STDERR "A short history of '"; if(scalar(@query_split)==2){print STDERR $query_split[0]."' and '".$query_split[1];}else{print STDERR $query_split[0]} print STDERR "' in '$project':\n";

print STDERR "--------------\n";
#
# input files
#
if($step eq "1" || $step eq "none" || $step eq "all"){

	if ( $ARGViddone_counter != scalar(@ARGV) ){
		print STDERR "[input] checking the input fasta files...\n";

		$did_found_a=0;
		$did_found_b=0;

		for(my $v = 0 ; $v < scalar @ARGV ; $v++){

			if($ARGViddone[$v]){next;}	
			if(! -e $ARGV[$v]){print STDERR "[input] ! WARNING ! did not found '".$ARGV[$v]."', proceeding anyway...\n"}
			elsif(! -r $ARGV[$v]){print STDERR "[input] ! WARNING ! cannot read '".$ARGV[$v]."', proceeding anyway...\n"}

			$result=`grep -nr '\\b$query_split[0]\\b' $ARGV[$v] 2>/dev/null`;
			$num_result=scalar split("\n",$result);

			if($num_result>0 && $did_found_a!=0){
				print STDERR "[input] ! WARNING ! found '".$query_split[0]."' again in ".$ARGV[$v].". Proteinortho can handle this and here the results for all are displayed ! \n";
				if($step ne "none"){chomp($result);print $result."\n";}
			}
			if($num_result==0){
				# did not found the query in this file ...
			}elsif($num_result==1 && $did_found_a==0){
				print STDERR "[input] found '".$query_split[0]."' in ".$ARGV[$v].". \n";
				if($step ne "none"){chomp($result);print $result."\n";}
				$did_found_a=1;
			}elsif($num_result>1){
				print STDERR "[input] ! ERROR ! found '".$query_split[0]."' multiple times in ".$ARGV[$v].". This seems dangerous, make sure there are no duplicates present in your fasta files ! \n";
				if($step ne "none"){chomp($result);print $result."\n";}
			}

			if(scalar(@query_split)==2){

				$result=`grep -nr '\\b$query_split[1]\\b' $ARGV[$v] 2>/dev/null`;
				$num_result=scalar split("\n",$result);

				if($num_result>0 && $did_found_b!=0){
					print STDERR "[input] ! WARNING ! found '".$query_split[1]."' again in ".$ARGV[$v].". Proteinortho can handle this and here the results for all are displayed ! \n";
					if($step ne "none"){chomp($result);print $result."\n";}
				}
				if($num_result==0){
					# did not found the query in this file ...
				}elsif($num_result==1 && $did_found_b==0){
					print STDERR "[input] found '".$query_split[1]."' in ".$ARGV[$v].". \n";
					if($step ne "none"){chomp($result);print $result."\n";}
					$did_found_b=1;
				}elsif($num_result>1){
					print STDERR "[input] ! ERROR ! found '".$query_split[1]."' multiple times in ".$ARGV[$v].". This seems dangerous, make sure there are no duplicates present in your fasta files ! \n";
					if($step ne "none"){chomp($result);print $result."\n";}
				}
			}
		}

		if(!$did_found_a){
			print STDERR "[input] ! ERROR ! did not found '".$query_split[0]."' in the provided fasta files...\n";
		}
		if(scalar(@query_split)==2 && !$did_found_b){
			print STDERR "[input] ! ERROR ! did not found '".$query_split[1]."' in the provided fasta files...\n";
		}
	}else{
		print STDERR "[input] input fasta files are not provided: skipping\n";
	}
}

print STDERR "--------------\n";
#
# step 2 blast-graph
# TODO the single blast graph ... 

if($step eq "2" || $step eq "none" || $step eq "all"){

	if(-d "proteinortho_cache_".$project){
		print STDERR "[all versus all blast] checking all blast files in 'proteinortho_cache_".$project."'...\n";


		$result=`grep -nr '\\b$query_split[0]\\b' proteinortho_cache_$project/*.vs.* 2>/dev/null`;
		$num_result=scalar split("\n",$result);

		print STDERR "[all versus all blast] the query '".$query_split[0]."' got $num_result hit(s).\n";
		if($step ne "none"){print $result;}

		if(scalar @query_split == 2){

			$result=`grep -nr '\\b$query_split[1]\\b' proteinortho_cache_$project/*.vs.* 2>/dev/null`;
			$num_result=scalar split("\n",$result);

			print STDERR "[all versus all blast] the query '".$query_split[1]."' got $num_result hit(s).\n";
			if($step ne "none"){print $result;}

			$result=`grep -nrP '(\\b$query_split[0]\\b\\t\\b$query_split[1]\\b)|(\\b$query_split[1]\\b\\t\\b$query_split[0]\\b)' proteinortho_cache_$project/*.vs.* 2>/dev/null`;
			$num_result=scalar split("\n",$result);

			print STDERR "[all versus all blast] both queries '".$query_split[0]."' and '".$query_split[1]."' have $num_result hit(s) with each other.\n";
			if($step ne "none"){print $result;}

			#print "grep -nr '(\\b$query_split[0]\\b\\t\\b$query_split[1]\\b)|(\\b$query_split[1]\\b\\t\\b$query_split[0]\\b)' proteinortho_cache_$project/*.vs.* 2>/dev/null";
		}

	}else{
		print STDERR "[all versus all blast] did not found temporary directory 'proteinortho_cache_".$project."/'. If you want to analyse all blast files please use '-keep' option...\n";
	}
print STDERR "--------------\n";

	if(-e $project.".blast-graph"){
		print STDERR "[all versus all blast] checking the blast-graph '$project.blast-graph' (reciprocal adaptive best hit graph)...\n";

		$result=`grep -nr '\\b$query_split[0]\\b' $project.blast-graph 2>/dev/null`;
		$num_result=scalar split("\n",$result);

		print STDERR "[all versus all blast] the query '".$query_split[0]."' got $num_result reciprocal hit(s).\n";
		if($step ne "none"){print $result;}

		if(scalar @query_split == 2){

			$result=`grep -nr '\\b$query_split[1]\\b' $project.blast-graph 2>/dev/null`;
			$num_result=scalar split("\n",$result);

			print STDERR "[all versus all blast] the query '".$query_split[1]."' got $num_result reciprocal hit(s).\n";
			if($step ne "none"){print $result;}

			$result=`grep -nrP '(\\b$query_split[0]\\b\\t\\b$query_split[1]\\b)|(\\b$query_split[1]\\b\\t\\b$query_split[0]\\b)' $project.blast-graph 2>/dev/null`;
			$num_result=scalar split("\n",$result);

			print STDERR "[all versus all blast] both queries '".$query_split[0]."' and '".$query_split[1]."' have $num_result reciprocal hit(s) with each other.\n";
			if($step ne "none"){print $result;}

			#print "grep -nr '(\\b$query_split[0]\\b\\t\\b$query_split[1]\\b)|(\\b$query_split[1]\\b\\t\\b$query_split[0]\\b)' proteinortho_cache_$project/*.vs.* 2>/dev/null";
		}

	}else{
		print STDERR "[all versus all blast] did not found the blast-graph '$project.blast-graph': skipping...\n";
	}
}

print STDERR "--------------\n";
#
# step 3 proteinortho
#
if($step eq "3" || $step eq "none" || $step eq "all"){

	if(-e $project.".proteinortho-graph"){
		print STDERR "[clustering] checking the proteinortho-graph '$project.proteinortho-graph' (result of the clustering)...\n";

		$result=`grep -nr '\\b$query_split[0]\\b' $project.proteinortho-graph 2>/dev/null`;
		$num_result=scalar split("\n",$result);

		print STDERR "[clustering] the query '".$query_split[0]."' got $num_result putative ortholog(s).\n";
		if($step ne "none"){print $result;}

		if(scalar @query_split == 2){

			$result=`grep -nr '\\b$query_split[1]\\b' $project.proteinortho-graph 2>/dev/null`;
			$num_result=scalar split("\n",$result);

			print STDERR "[clustering] the query '".$query_split[1]."' got $num_result putative ortholog(s).\n";
			if($step ne "none"){print $result;}

			$result=`grep -nrP '(\\b$query_split[0]\\b\\t\\b$query_split[1]\\b)|(\\b$query_split[1]\\b\\t\\b$query_split[0]\\b)' $project.proteinortho-graph 2>/dev/null`;
			$num_result=scalar split("\n",$result);

			if($num_result!=0){
				print STDERR "[clustering] both queries '".$query_split[0]."' and '".$query_split[1]."' are putative ortholog(s).\n";
				if($step ne "none"){print $result;}
			}else{
				print STDERR "[clustering] both queries '".$query_split[0]."' and '".$query_split[1]."' are NOT putative ortholog(s).\n";
			}
			
			#print "grep -nr '(\\b$query_split[0]\\b\\t\\b$query_split[1]\\b)|(\\b$query_split[1]\\b\\t\\b$query_split[0]\\b)' proteinortho_cache_$project/*.vs.* 2>/dev/null";
		}

	}else{
		print STDERR "[clustering] did not found the proteinortho-graph '$project.proteinortho-graph': skipping...\n";
	}
print STDERR "--------------\n";


	if(-e $project.".proteinortho.tsv"){
		print STDERR "[clustering] checking the proteinortho.tsv file '$project.proteinortho.tsv' (result of the clustering)...\n";

		$result=`grep -nr '\\b$query_split[0]\\b' $project.proteinortho.tsv 2>/dev/null`;
		$num_result=scalar split("\n",$result);

		print STDERR "[clustering] the query '".$query_split[0]."' is part of $num_result group(s) of putative orthologs.\n";
		if($step ne "none"){print $result;}

		if(scalar @query_split == 2){

			$result=`grep -nr '\\b$query_split[1]\\b' $project.proteinortho.tsv 2>/dev/null`;
			$num_result=scalar split("\n",$result);

			print STDERR "[clustering] the query '".$query_split[1]."' is part of $num_result group(s) of putative orthologs.\n";
			if($step ne "none"){print $result;}

			$result=`grep -nrP '(\\b$query_split[0]\\b.*\\b$query_split[1]\\b)|(\\b$query_split[1]\\b.*\\b$query_split[0]\\b)' $project.proteinortho.tsv 2>/dev/null`;
			$num_result=scalar split("\n",$result);

			print STDERR "[clustering] both queries '".$query_split[0]."' and '".$query_split[1]."' have $num_result group(s) in common.\n";
			if($step ne "none"){print $result;}

			#print "grep -nr '(\\b$query_split[0]\\b\\t\\b$query_split[1]\\b)|(\\b$query_split[1]\\b\\t\\b$query_split[0]\\b)' proteinortho_cache_$project/*.vs.* 2>/dev/null";
		}

	}else{
		print STDERR "[clustering] did not found the proteinortho.tsv '$project.proteinortho.tsv': skipping...\n";
	}
}
